import sys
import warnings
import pandas as pd
import yaml
from yaml.loader import SafeLoader

import seaborn as sns

sns.set_context('paper')

warnings.filterwarnings('ignore')

from Topyfic.utilsAnalyseModel import GSEA, functional_enrichment_analysis


class Topic:
    """
    A class saved topic along with other useful information

    :param topic_id: ID of topic which is unique
    :type topic_id: str
    :param topic_name: name of the topic (default: topic_id)
    :type topic_name: str
    :param topic_gene_weights: dataframe that contains weights of the topics for each genes
    :type topic_gene_weights: pandas dataframe
    :param gene_information: dataframe that contains information of genes i.e gene biotype
    :type gene_information: pandas dataframe
    :param topic_information: dataframe that contains information of genes i.e cell state / cell type
    :type topic_information: pandas dataframe
    """

    def __init__(self,
                 topic_id,
                 topic_name=None,
                 topic_gene_weights=None,
                 gene_information=None,
                 topic_information=None):

        self.id = topic_id

        if topic_name is not None:
            self.name = topic_name
        else:
            self.name = topic_id

        self.gene_weights = topic_gene_weights

        if gene_information is None:
            self.gene_information = pd.DataFrame(index=topic_gene_weights.index)
        else:
            self.gene_information = gene_information

        if topic_information is None:
            self.topic_information = pd.DataFrame(index=topic_gene_weights.columns)
        else:
            self.topic_information = topic_information

    def update_gene_information(self, gene_information):
        """
        update/add genes information for each topics

        :param gene_information: dataframe contains genes information we would like to add/update (the index should be same as an index of gene_information in class)
        :type gene_information: pandas dataframe
        """
        same_columns = self.gene_information.columns.intersection(gene_information.columns)
        self.gene_information.drop(same_columns, axis=1, inplace=True)
        self.gene_information = pd.concat([self.gene_information, gene_information], axis=1)

    def functional_enrichment_analysis(self,
                                       type,
                                       organism,
                                       sets=None,
                                       p_value=0.05,
                                       file_format="pdf",
                                       file_name="functional_enrichment_analysis"):
        """
        Doing functional enrichment analysis including GO, KEGG and REACTOME

        :param type: indicate the type of databases which it should be one of "GO", "REACTOME"
        :type type: str
        :param organism: name of the organ you want to do functional enrichment analysis
        :type organism: str
        :param sets: str, list, tuple of Enrichr Library name(s). (you can add any Enrichr Libraries from here: https://maayanlab.cloud/Enrichr/#stats) only need to fill if the type is GO
        :type sets: str, list, tuple
        :param p_value: Defines the pValue threshold. (default: 0.05)
        :type p_value: float
        :param file_format: indicate the format of plot (default: pdf)
        :type file_format: str
        :param file_name: name and path of the plot use for save (default: gene_composition)
        :type file_name: str
        """
        gene_weights = self.gene_weights.copy(deep=True)
        gene_weights = gene_weights[gene_weights > gene_weights.min()]
        genes = gene_weights.dropna().index.tolist()

        functional_enrichment_analysis(gene_list=genes,
                                       type=type,
                                       organism=organism,
                                       sets=sets,
                                       p_value=p_value,
                                       file_format=file_format,
                                       file_name=file_name)

    def GSEA(self,
             gene_sets='GO_Biological_Process_2021',
             p_value=0.05,
             table=True,
             plot=True,
             file_format="pdf",
             file_name="GSEA",
             **kwargs):
        """
        Doing Gene Set Enrichment Analysis on based on the topic weights using GSEAPY package.

        :param gene_sets: Enrichr Library name or .gmt gene sets file or dict of gene sets. (you can add any Enrichr Libraries from here: https://maayanlab.cloud/Enrichr/#stats)
        :type gene_sets: str, list, tuple
        :param p_value: Defines the pValue threshold for plotting. (default: 0.05)
        :type p_value: float
        :param table: indicate if you want to save all GO terms that passed the threshold as a table (default: True)
        :type table: bool
        :param plot: indicate if you want to plot all GO terms that passed the threshold (default: True)
        :type plot: bool
        :param file_format: indicate the format of plot (default: pdf)
        :type file_format: str
        :param file_name: name and path of the plot use for save (default: gene_composition)
        :type file_name: str
        :param kwargs: Argument to pass to gseapy.prerank(). more info: https://gseapy.readthedocs.io/en/latest/run.html?highlight=gp.prerank#gseapy.prerank

        :return: dataframe contains these columns: Term: gene set name, ES: enrichment score, NES: normalized enrichment score, NOM p-val:  Nominal p-value (from the null distribution of the gene set, FDR q-val: FDR qvalue (adjusted False Discory Rate), FWER p-val: Family wise error rate p-values, Tag %: Percent of gene set before running enrichment peak (ES), Gene %: Percent of gene list before running enrichment peak (ES), Lead_genes: leading edge genes (gene hits before running enrichment peak)
        :rtype: pandas dataframe
        """

        gene_weights = self.gene_weights.copy(deep=True)
        gene_weights.index = gene_weights.index.str.upper()
        gene_weights.sort_values([self.id], ascending=False, inplace=True)
        gene_weights = gene_weights[gene_weights > gene_weights.min()]
        gene_weights.dropna(inplace=True)

        if gene_weights.shape[0] == 1:
            return

        GSEA_df = GSEA(gene_list=gene_weights,
                       gene_sets=gene_sets,
                       p_value=p_value,
                       table=table,
                       plot=plot,
                       file_format=file_format,
                       file_name=file_name,
                       **kwargs)

        return GSEA_df

    def gene_weight_variance(self, save=True):
        """
        calculate the gene weight variance

        :param save: added as an information to the Topic (default: True)
        :type save: bool

        :return: Gene weight variance for given topic
        :rtype: float
        """
        variance = self.gene_weights.var().tolist()[0]

        if save:
            self.topic_information['variance'] = self.gene_weights.var().tolist()[0]

        return print(f"Gene weight variance for given topic is {variance}")

    def write_topic_yaml(self, topic_id=None, model_yaml_path="model.yaml", topic_yaml_path="topic.yaml", save=True):
        """
        write topic in YAML format

        :param topic_id: unique topic ID (default is topic ID)
        :type topic_id: str
        :param model_yaml_path: model yaml path that has information about the dataset you use
        :type model_yaml_path: str
        :param topic_yaml_path: path that you use to save topic
        :type topic_yaml_path: str
        :param save: indicate if you want to save yaml file (True) or just show them (Fasle) (default: True)
        :type save: bool
        """

        # Open the file and load the file
        with open(model_yaml_path) as f:
            model_yaml = yaml.load(f, Loader=SafeLoader)

        if topic_id is None:
            topic_id = self.id

        if topic_id not in model_yaml['Topic file_name(s)']:
            sys.exit("Topic_id is not in model YAML file!")

        topic_yaml = {'Topic ID': topic_id,
                      'Gene weights': self.gene_weights.to_dict()[self.id],
                      'Gene information': self.gene_information.to_dict(),
                      'Topic information': self.topic_information.T.to_dict()[self.id]}

        if save:
            file = open(topic_yaml_path, "w")
            yaml.dump(topic_yaml, file, default_flow_style=False)
            file.close()
        else:
            yaml_string = yaml.dump(topic_yaml)
            print("The Topic YAML is:")
            print(yaml_string)



import sys
import random
import pandas as pd
import numpy as np
import anndata
import scanpy as sc
import warnings
import joblib
import os
from multiprocessing import Pool
from itertools import repeat
import pickle
from sklearn.decomposition import LatentDirichletAllocation
from scipy import stats as st
from scipy.spatial import distance
from sklearn.metrics.pairwise import cosine_similarity
import scanpy.external as sce
import networkx as nx
import math
import matplotlib.pyplot as plt
import seaborn as sns
import gseapy as gp
from gseapy.plot import dotplot
from gseapy import gseaplot
from reactome2py import analysis
from adjustText import adjust_text
import umap
import obonet
import plotly.express as px

warnings.filterwarnings("ignore")


def compare_topModels(topModels,
                      comparison_method="Jensen–Shannon divergence",
                      output_type='graph',
                      threshold=0.8,
                      topModels_color=None,
                      topModels_label=None,
                      ignore_genes=True,
                      save=False,
                      plot_show=True,
                      figsize=None,
                      plot_format="pdf",
                      file_name="compare_topics"):
    """
    compare topModels using topic gene weights

    :param topModels: list of topModel class you want to compare to each other
    :type topModels: list of TopModel class
    :param comparison_method: indicate the method you want to use for comparing topics. if you used Jensen–Shannon, we show -log2 (options: pearson correlation, spearman correlation, Jensen–Shannon divergence, cosine similarity)
    :type comparison_method: str
    :param output_type: indicate the type of output you want. graph: plot as a graph, heatmap: plot as a heatmap, table: table contains correlation. Note: if you want to plot Jensen–Shannon divergence as a graph, we convert the values to be at the -log2(), so you need to take that account for defining threshold
    :type output_type: str
    :param threshold: only apply when you choose circular which only show correlation above that
    :type threshold: float
    :param topModels_color: dictionary of colors mapping each topics to each color (default: blue)
    :type topModels_color: dict
    :param topModels_label: dictionary of label mapping each topics to each label
    :type topModels_label: dict
    :param ignore_genes: idicate how to behave to the genes that only represent in one the topics. "True" means it's gonna ignore those genes and "False" means it's gonna asumne the weights are zero for those genes that we don't have any weights in one of the mouse models
    :type ignore_genes: bool
    :param save: indicate if you want to save the plot or not (default: True)
    :type save: bool
    :param plot_show: indicate if you want to show the plot or not (default: True)
    :type plot_show: bool
    :param figsize: indicate the size of plot (default: (10 * (len(category) + 1), 10))
    :type figsize: tuple of int
    :param plot_format: indicate the format of plot (default: pdf)
    :type plot_format: str
    :param file_name: name and path of the plot use for save (default: compare_topics)
    :type file_name: str

    :return: table contains correlation between topics only when table is choose and save is False
    :rtype: pandas dataframe
    """
    if output_type not in ['graph', 'heatmap', 'table']:
        sys.exit("output_type is not valid! it should be one of 'graph', 'heatmap', or 'table'")

    if comparison_method not in ['spearman correlation', 'pearson correlation', 'Jensen–Shannon divergence', 'cosine similarity']:
        sys.exit("comparison_method is not valid! it should be one of 'spearman correlation', 'pearson correlation', or 'Jensen–Shannon divergence'")

    names = [topModel.name for topModel in topModels]
    if len(names) != len(set(names)):
        sys.exit("Name of the TopModels should be unique!")

    all_gene_weights = None

    for topModel in topModels:
        gene_weights = topModel.get_gene_weights()
        if all_gene_weights is None:
            all_gene_weights = gene_weights
        else:
            all_gene_weights = pd.concat([gene_weights, all_gene_weights], axis=1)

    corrs = pd.DataFrame(index=all_gene_weights.columns,
                         columns=all_gene_weights.columns,
                         dtype='float64')

    for d1 in all_gene_weights.columns.tolist():
        for d2 in all_gene_weights.columns.tolist():
            if d1 == d2:
                corrs.at[d1, d2] = 1
                continue
            a = all_gene_weights[[d1, d2]]
            if ignore_genes:
                a.dropna(axis=0, how='any', inplace=True)
            else:
                a.dropna(axis=0, how='all', inplace=True)
                a.fillna(0, inplace=True)

            a = a[np.logical_or(a[d1] > a[d1].min(), a[d2] > a[d2].min())]
            a = a / a.sum()
            if comparison_method == "Jensen–Shannon divergence":
                JSd = distance.jensenshannon(a[d1].tolist(), a[d2].tolist())
                corrs.loc[d1, d2] = JSd * JSd
            elif comparison_method == "pearson correlation":
                corr = st.pearsonr(a[d1].tolist(), a[d2].tolist())
                corrs.at[d1, d2] = corr[0]
            elif comparison_method == "spearman correlation":
                corr = st.spearmanr(a[d1].tolist(), a[d2].tolist())
                corrs.at[d1, d2] = corr[0]
            elif comparison_method == 'cosine similarity':
                corr = distance.cosine(a[d1].tolist(), a[d2].tolist())
                corrs.at[d1, d2] = 1 - corr
    if comparison_method == "Jensen–Shannon divergence":
        corrs = corrs.applymap(math.log2)
        corrs = corrs * -1

    if output_type == 'table':
        if save:
            corrs.to_csv(f"{file_name}.csv")
        else:
            return corrs

    if output_type == 'heatmap':
        if topModels_label is not None:
            corrs.rename(columns=topModels_label, index=topModels_label, inplace=True)
        sns.clustermap(corrs,
                       figsize=figsize)
        if save:
            plt.savefig(f"{file_name}.{plot_format}")
        if plot_show:
            plt.show()
        else:
            plt.close()

        return

    if output_type == 'graph':
        np.fill_diagonal(corrs.values, 0)
        corrs[corrs < threshold] = np.nan
        res = corrs.stack()
        res = pd.DataFrame(res)
        res.reset_index(inplace=True)
        res.columns = ['source', 'dest', 'weight']
        res['weight'] = res['weight'].astype(float).round(decimals=2)
        res['source_label'] = res['source']
        res['dest_label'] = res['dest']
        res['source_color'] = res['source']
        res['dest_color'] = res['dest']

        if topModels_label is not None:
            res['source_label'].replace(topModels_label, inplace=True)
            res['dest_label'].replace(topModels_label, inplace=True)
        if topModels_color is None:
            res['source_color'] = "blue"
            res['dest_color'] = "blue"
        else:
            res['source_color'].replace(topModels_color, inplace=True)
            res['dest_color'].replace(topModels_color, inplace=True)

        G = nx.Graph()
        for i in range(res.shape[0]):
            G.add_node(res.source_label[i], color=res.source_color[i])
            G.add_node(res.dest_label[i], color=res.dest_color[i])
            G.add_edge(res.source_label[i], res.dest_label[i], weight=res.weight[i])

        connected_components = sorted(nx.connected_components(G), key=len, reverse=True)

        if figsize is None:
            figsize = (len(connected_components) * 2, len(connected_components) * 2)

        nrows = math.ceil(math.sqrt(len(connected_components)))
        ncols = math.ceil(len(connected_components) / nrows)
        fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, facecolor='white')

        i = 0
        for connected_component in connected_components:
            g_connected_component = G.subgraph(connected_component)
            nodePos = nx.spring_layout(g_connected_component)

            edge_labels = nx.get_edge_attributes(g_connected_component, "weight")

            node_color = nx.get_node_attributes(g_connected_component, "color").values()
            weights = nx.get_edge_attributes(g_connected_component, 'weight').values()

            if len(connected_components) == 1:
                ax = axs
            elif ncols == 1:
                ax = axs[int(i / ncols)]
            else:
                ax = axs[int(i / ncols), i % ncols]

            nx.draw_networkx(g_connected_component,
                             pos=nodePos,
                             width=list(weights),
                             with_labels=True,
                             node_color=list(node_color),
                             font_size=8,
                             node_size=500,
                             font_family='Arial',
                             ax=ax)

            nx.draw_networkx_edge_labels(g_connected_component,
                                         nodePos,
                                         edge_labels=edge_labels,
                                         font_size=7,
                                         ax=ax)

            i += 1

        if len(connected_components) == 1:
            ax.axis('off')
        else:
            [axi.axis('off') for axi in axs.ravel()]
        plt.tight_layout()
        if save:
            plt.savefig(f"{file_name}.{plot_format}")
        if plot_show:
            plt.show()
        else:
            plt.close()

        return axs


def MA_plot(topic1,
            topic2,
            size=None,
            pseudocount=1,
            threshold=1,
            cutoff=2.0,
            consistency_correction=1.4826,
            #topN=None,
            labels=None,
            save=True,
            show=True,
            file_format="pdf",
            file_name="MA_plot"):
    """
        plot MA based on the gene weights on given topics

        :param topic1: gene weight of first topic to be compared
        :type topic1: pandas.series
        :param topic2: gene weight of second topic to be compared
        :type topic2: pandas.series
        :param size: table contains size of dot for each genes (genes are index)
        :type size: pandas dataframe
        :param pseudocount: pseudocount that you want to add (default: 1)
        :type pseudocount: float
        :param threshold: threshold to filter genes based on A values (default: 1)
        :type threshold: float
        :param cutoff: cutoff for categorized genes by modified z-score (default: 2)
        :type cutoff: float
        :param consistency_correction: the factor converts the MAD to the standard deviation for a given distribution. The default value (1.4826) is the conversion factor if the underlying data is normally distributed
        :type consistency_correction: float
        :param topN: number of genes to be consider for calculating z-score based on the A value (if it's none is gonna be avarage of # genes in both topics with weights above threshold
        :type topN: int
        :param labels: list of gene names wish to show in MA-plot
        :type labels: list
        :param save: indicate if you want to save the plot or not (default: True)
        :type save: bool
        :param show: indicate if you want to show the plot or not (default: True)
        :type show: bool
        :param file_format: indicate the format of plot (default: pdf)
        :type file_format: str
        :param file_name: name and path of the plot use for save (default: MA_plot)
        :type file_name: str

        :return: return M and A values
        """
    topic1 += pseudocount
    topic2 += pseudocount

    A = (np.log2(topic1) + np.log2(topic2)) / 2
    M = np.log2(topic1) - np.log2(topic2)

    # if topN is None:
    #     len_topic1 = sum(topic1 > threshold)
    #     len_topic2 = sum(topic2 > threshold)
    #     topN = round((len_topic1 + len_topic2) / 2)

    gene_zscore = pd.concat([A, M], axis=1)
    gene_zscore.columns = ["A", "M"]
    gene_zscore = gene_zscore[gene_zscore.A > threshold]

    if size is not None:
        size = size.loc[gene_zscore.index, :]
        gene_zscore = pd.concat([gene_zscore, size], axis=1)
        gene_zscore.columns = ["A", "M", "size"]

    gene_zscore.sort_values('A', ascending=False, inplace=True)
    #gene_zscore = gene_zscore.iloc[:topN, :]

    gene_zscore['mod_zscore'], mad = modified_zscore(gene_zscore['M'],
                                                     consistency_correction=consistency_correction)

    if gene_zscore.shape[0] == 0:
        print("there is no genes that pass the threshold!")
        return gene_zscore
    plot_df = gene_zscore.copy(deep=True)
    plot_df.mod_zscore = plot_df.mod_zscore.abs()
    plot_df.mod_zscore[plot_df.mod_zscore > cutoff] = cutoff
    plot_df.mod_zscore[plot_df.mod_zscore < cutoff] = 0
    plot_df.mod_zscore.replace(float('-inf'), -1*cutoff)
    plot_df.mod_zscore.replace(float('inf'), cutoff)
    plot_df.mod_zscore.fillna(0, inplace=True)
    plot_df.mod_zscore = plot_df.mod_zscore.astype(str)
    plot_df.mod_zscore[plot_df.mod_zscore == str(cutoff)] = f'> {cutoff}'
    plot_df.mod_zscore[plot_df.mod_zscore == '0.0'] = f'< {cutoff}'

    if labels is None:
        plot_df['label'] = ""
    else:
        plot_df['label'] = plot_df.index.tolist()
        plot_df.label[~plot_df.label.isin(labels)] = ""

    y = plot_df.M.median()
    ymin = y - consistency_correction * mad * cutoff
    ymax = y + consistency_correction * mad * cutoff
    xmin = round(plot_df.A.min()) - 1
    xmax = round(plot_df.A.max())
    if size is None:
        plot_df.columns = ["A", "M", "abs(mod_Zscore)", "label"]
    else:
        plot_df.columns = ["A", "M", "#topics GW >= 1", "abs(mod_Zscore)", "label"]

    color_palette = {f'> {cutoff}': "orchid",
                     f'< {cutoff}': "royalblue"}
    markers = {f'> {cutoff}': "o",
               f'< {cutoff}': "s"}

    if size is None:
        sns.scatterplot(data=plot_df, x="A", y="M", style="abs(mod_Zscore)", hue="abs(mod_Zscore)",
                        markers=markers, palette=color_palette, linewidth=0.1)
    else:
        sns.scatterplot(data=plot_df, x="A", y="M", style="abs(mod_Zscore)", hue="#topics GW >= 1",
                        linewidth=0.1, markers=markers)

    if labels is not None:
        texts = []
        for label in plot_df.label.unique():
            if label != "":
                texts.append(plt.text(plot_df.A[label], plot_df.M[label], label, horizontalalignment='left'))

        adjust_text(texts, arrowprops=dict(arrowstyle="-", color='black', lw=0.5))

    plt.legend(loc='center left', bbox_to_anchor=(1.05, 0.5), ncol=1)
    plt.hlines(y=y, xmin=xmin, xmax=xmax, colors="red")
    plt.hlines(y=ymin, xmin=xmin, xmax=xmax, colors="orange", linestyles='--')
    plt.hlines(y=ymax, xmin=xmin, xmax=xmax, colors="orange", linestyles='--')

    if save:
        plt.savefig(f"{file_name}.{file_format}")
    if show:
        plt.show()
    else:
        plt.close()

    return gene_zscore


def modified_zscore(data, consistency_correction=1.4826):
    """
    Returns the modified z score and Median Absolute Deviation (MAD) from the scores in data.
    The consistency_correction factor converts the MAD to the standard deviation for a given
    distribution. The default value (1.4826) is the conversion factor if the underlying data
    is normally distributed
    """

    median = np.median(data)

    deviation_from_med = np.array(data) - median

    mad = np.median(np.abs(deviation_from_med))
    mod_zscore = deviation_from_med / (consistency_correction * mad)

    return mod_zscore, mad


def functional_enrichment_analysis(gene_list,
                                   type,
                                   organism,
                                   sets=None,
                                   p_value=0.05,
                                   file_format="pdf",
                                   file_name="functional_enrichment_analysis"):
    """
    Doing functional enrichment analysis including GO, KEGG and REACTOME

    :param gene_list: list of gene name
    :type gene_list: list
    :param type: indicate the type of databases which it should be one of "GO", "REACTOME"
    :type type: str
    :param organism: name of the organ you want to do functional enrichment analysis
    :type organism: str
    :param sets: str, list, tuple of Enrichr Library name(s). (you can add any Enrichr Libraries from here: https://maayanlab.cloud/Enrichr/#stats) only need to fill if the type is GO
    :type sets: str, list, tuple
    :param p_value: Defines the pValue threshold. (default: 0.05)
    :type p_value: float
    :param file_format: indicate the format of plot (default: pdf)
    :type file_format: str
    :param file_name: name and path of the plot use for save (default: gene_composition)
    :type file_name: str
    """

    if type not in ["GO", "REACTOME"]:
        sys.exit("Type is not valid! it should be one of them GO, KEGG, REACTOME")

    if type == "GO" and sets is None:
        sets = ["GO_Biological_Process_2021"]
    elif type == "KEGG" and sets is None:
        sets = ["KEGG_2016"]

    if type in ["GO", "KEGG"]:
        enr = gp.enrichr(gene_list=gene_list,
                         gene_sets=sets,
                         organism=organism,
                         outdir=f"{file_name}",
                         cutoff=p_value)
        dotplot(enr.res2d,
                title=f"Gene ontology",
                cmap='viridis_r',
                cutoff=p_value,
                ofname=f"{file_name}.{file_format}")
    else:
        numGeneModule = len(gene_list)
        genes = ",".join(gene_list)
        result = analysis.identifiers(ids=genes,
                                      species=organism,
                                      p_value=str(p_value))
        token = result['summary']['token']
        analysis.report(token,
                        path=f"{file_name}/",
                        file=f"{file_name}.{file_format}",
                        number='50',
                        species=organism)
        token_result = analysis.token(token,
                                      species=organism,
                                      p_value=str(p_value))

        print(
            f"{numGeneModule - token_result['identifiersNotFound']} out of {numGeneModule} identifiers in the sample were found in Reactome.")
        print(
            f"{token_result['resourceSummary'][0]['pathways']} pathways were hit by at least one of them, which {len(token_result['pathways'])} of them have p-value more than {p_value}.")
        print(f"Report was saved {file_name}!")
        print(f"For more information please visit https://reactome.org/PathwayBrowser/#/DTAB=AN&ANALYSIS={token}")


def GSEA(gene_list,
         gene_sets='GO_Biological_Process_2021',
         p_value=0.05,
         table=True,
         plot=True,
         file_format="pdf",
         file_name="GSEA",
         **kwargs):
    """
    Doing Gene Set Enrichment Analysis on based on the topic weights using GSEAPY package.

    :param gene_list: pandas series with index as a gene names and their ranks/weights as value
    :type gene_list: pandas series
    :param gene_sets: Enrichr Library name or .gmt gene sets file or dict of gene sets. (you can add any Enrichr Libraries from here: https://maayanlab.cloud/Enrichr/#stats)
    :type gene_sets: str, list, tuple
    :param p_value: Defines the pValue threshold for plotting. (default: 0.05)
    :type p_value: float
    :param table: indicate if you want to save all GO terms that passed the threshold as a table (default: True)
    :type table: bool
    :param plot: indicate if you want to plot all GO terms that passed the threshold (default: True)
    :type plot: bool
    :param file_format: indicate the format of plot (default: pdf)
    :type file_format: str
    :param file_name: name and path of the plot use for save (default: gene_composition)
    :type file_name: str
    :param kwargs: Argument to pass to gseapy.prerank(). more info: https://gseapy.readthedocs.io/en/latest/run.html?highlight=gp.prerank#gseapy.prerank

    :return: dataframe contains these columns: Term: gene set name, ES: enrichment score, NES: normalized enrichment score, NOM p-val:  Nominal p-value (from the null distribution of the gene set, FDR q-val: FDR qvalue (adjusted False Discory Rate), FWER p-val: Family wise error rate p-values, Tag %: Percent of gene set before running enrichment peak (ES), Gene %: Percent of gene list before running enrichment peak (ES), Lead_genes: leading edge genes (gene hits before running enrichment peak)
    :rtype: pandas dataframe
    """

    gene_list.index = gene_list.index.str.upper()

    pre_res = gp.prerank(rnk=gene_list,
                         gene_sets=gene_sets,
                         format=file_format,
                         no_plot=~plot,
                         **kwargs)

    pre_res.res2d.sort_values(["NOM p-val"], inplace=True)
    pre_res.res2d.drop(["Name"], axis=1, inplace=True)

    if table:
        pre_res.res2d.to_csv(f"{file_name}.csv")

    if plot:
        res = pre_res.res2d.copy(deep=True)
        res = res[res['NOM p-val'] < p_value]
        for term in res.Term:
            name = term.split("(GO:")[1][:-1]
            gseaplot(rank_metric=pre_res.ranking,
                     term=term,
                     **pre_res.results[term],
                     ofname=f"{file_name}_GO_{name}.{file_format}")

    return pre_res.res2d


def summarize_GO_Term(GO_terms,
                      p_value=0.05,
                      file_format="html",
                      file_name="GO_sum"):
    """
    Summarize long, unintelligible lists of GO terms by finding a representative subset of the terms showing more unique (child) Go terms
    We suggest save it as html since it's gonna be plot by plotly so you can take an advantage of using plotly

    :param GO_terms: Dataframe contains results of gene ontology analysis performs by GSEAPY (https://gseapy.readthedocs.io/en/latest/index.html)
    :type GO_terms: pandas dataframe
    :param p_value: Defines the pValue threshold for plotting. (default: 0.05)
    :type p_value: float
    :param file_format: indicate the format of plot (default: html)
    :type file_format: str
    :param file_name: name and path of the plot use for save (default: gene_composition)
    :type file_name: str

    :return: dataframe used to plot the results
    :rtype: pandas dataframe
    """

    tmp = GO_terms.Term.str.split(pat="(", expand=True)
    GO_terms.Term = tmp[0]
    GO_terms['GO_id'] = tmp[1].str.split(pat=")", expand=True)[0]

    genes = ";".join(GO_terms['Lead_genes'].tolist())
    genes = genes.split(";")
    genes = list(set(genes))

    go_genes = pd.DataFrame(0, index=GO_terms.GO_id, columns=genes)
    for i in range(go_genes.shape[0]):
        tmp = GO_terms.Lead_genes[GO_terms.GO_id == go_genes.index[i]].tolist()[0]
        tmp = tmp.split(";")
        go_genes.loc[go_genes.index[i], tmp] = 1

    go_go_cosine_similarity = pd.DataFrame(cosine_similarity(go_genes),
                                           index=GO_terms.GO_id,
                                           columns=GO_terms.GO_id,
                                           dtype='float64')

    fit = umap.UMAP()

    # principalComponents = pca.fit_transform(go_genes)
    u = fit.fit_transform(go_genes)

    df = pd.DataFrame(data=u,
                      index=go_genes.index.tolist(),
                      columns=['umap1', 'umap2'],
                      dtype=float)

    GO_terms.index = GO_terms.GO_id.values
    df = pd.concat([df, GO_terms], axis=1)
    df['Gene %'] = df['Gene %'].str.replace('%', 'e-2').astype(float)

    url = 'http://purl.obolibrary.org/obo/go.obo'
    graph = obonet.read_obo(url)

    remove_goid = []
    for source in go_go_cosine_similarity.index.tolist():
        for target in go_go_cosine_similarity.index.tolist():
            if graph.has_node(source) and graph.has_node(target) and nx.has_path(graph, source=source,
                                                                                 target=target) and source != target:
                remove_goid.append(target)
                # print(nx.shortest_path(graph, source=source, target=target))
    remove_goid = list(set(remove_goid))
    # df = go_go_cosine_similarity.drop(remove_goid, axis=0)
    df.drop(remove_goid, axis=0, inplace=True)

    df = df[df['NOM p-val'] < p_value]

    fig = px.scatter(df,
                     x="umap1",
                     y="umap2",
                     size="Gene %",
                     color="NOM p-val",
                     hover_data=['GO_id', 'Term'],
                     color_continuous_scale=px.colors.sequential.Viridis_r)
    fig.show()
    fig.write_html(f"{file_name}.{file_format}")

    return df




import sys
import random
import pandas as pd
import numpy as np
import anndata
import scanpy as sc
import warnings
import joblib
import os
from multiprocessing import Pool
from itertools import repeat
import pickle
from sklearn.decomposition import LatentDirichletAllocation, NMF
from scipy import stats as st
import scanpy.external as sce
import networkx as nx
import math
import matplotlib.pyplot as plt
import seaborn as sns
import gseapy as gp
from gseapy.plot import dotplot
from gseapy import gseaplot
from reactome2py import analysis
import yaml
from yaml.loader import SafeLoader
import h5py

from Topyfic.train import Train
from Topyfic.analysis import Analysis
from Topyfic.topModel import TopModel
from Topyfic.topic import Topic

warnings.filterwarnings("ignore")


def train_model(name,
                data,
                k,
                n_runs=100,
                random_state_range=None,
                n_thread=5,
                save_path=""):
    """
    Training model and save it

    :param name: name of the Train class
    :type name: str
    :param k: number of topics to learn one LDA model using sklearn package (default: 50)
    :type k: int
    :param n_runs: number of run to define rLDA model (default: 100)
    :type n_runs: int
    :param random_state_range: list of random state, we used to run LDA models (default: range(n_runs))
    :type random_state_range: list of int
    :param data: data embedded in anndata format use to train LDA model
    :type data: anndata
    :param n_thread: number of threads you used to learn LDA models (default=5)
    :type n_thread: int
    :param save_path: directory you want to use to save pickle file (default is saving near script)
    :type save_path: str

    """
    train = Train(name=name,
                  k=k,
                  n_runs=n_runs,
                  random_state_range=random_state_range)
    train.run_LDA_models(data, n_thread=n_thread)
    train.save_train(save_path=save_path)


def make_topModel(trains,
                  data,
                  n_top_genes=50,
                  resolution=1,
                  file_format="pdf",
                  save_path=""):
    """
    Creating topModel base on train data and save it along with clustering information

    :param trains: list of train class
    :type trains: list of Train
    :param data: expression data embedded in anndata format along with cells and genes/region information
    :type data: anndata
    :param n_top_genes: Number of highly-variable genes to keep (default: 50)
    :type n_top_genes: int
    :param resolution: A parameter value controlling the coarseness of the clustering. Higher values lead to more clusters. (default: 1)
    :type resolution: int
    :param file_format: indicate the format of plot (default: pdf)
    :type file_format: str
    :param save_path: directory you want to use to save pickle file (default is saving near script)
    :type save_path: str

    """
    top_model, clustering = calculate_leiden_clustering(trains=trains,
                                                        data=data,
                                                        n_top_genes=n_top_genes,
                                                        resolution=resolution,
                                                        file_format=file_format)
    clustering.to_csv(f"{save_path}/clustering.csv")
    top_model.save_topModel(save_path=save_path)


def make_analysis_class(top_model,
                        data,
                        colors_topics=None,
                        save_path=""):
    """
    Creating Analysis object

    :param top_model: top model that used for analysing topics, gene weights compositions and calculate cell participation
    :type top_model: TopModel
    :param data: processed expression data along with cells and genes/region information
    :type data: anndata
    :param colors_topics: dataframe that mapped colored to topics
    :type colors_topics: pandas dataframe
    :param save_path: directory you want to use to save pickle file (default is saving near script)
    :type save_path: str
    """
    analysis_top_model = Analysis(Top_model=top_model, colors_topics=colors_topics)
    analysis_top_model.calculate_cell_participation(data=data)
    analysis_top_model.save_analysis(save_path=save_path)


def subset_data(data, keep, loc='var'):
    """
    Subsetting data

    :param data: data we want to subset
    :type data: anndata
    :param keep: values in the obs/var_names
    :type keep: list
    :param loc: subsetting in which direction (default: 'var')

    :return: data we want to keep
    :rtype: anndata
    """
    if loc == 'var':
        data = data[:, keep]
    elif loc == 'obs':
        data = data[keep, :]
    else:
        sys.exit("loc is not correct! it should be 'obs' or 'var'")

    return data


def calculate_leiden_clustering(trains,
                                data,
                                n_top_genes=None,
                                resolution=1,
                                max_iter_harmony=10,
                                min_cell_participation=None,
                                file_format="pdf"):
    """
    Do leiden clustering w/o harmony base on number of assays you have and then remove low participation topics

    :param trains: list of train class
    :type trains: list of Train
    :param data: gene-count data with cells and genes information
    :type data: anndata
    :param n_top_genes: Number of highly-variable genes to keep (default: 50)
    :type n_top_genes: int
    :param resolution: A parameter value controlling the coarseness of the clustering. Higher values lead to more clusters. (default: 1)
    :type resolution: int
    :param max_iter_harmony: number of iteration for running harmony (default: 10)
    :type max_iter_harmony: int
    :param min_cell_participation: minimum cell participation across for each topics to keep them, when is None, it will keep topics with cell participation more than 1% of #cells (#cells / 100)
    :type min_cell_participation: float
    :param file_format: indicate the format of plot (default: pdf)
    :type file_format: str

    :return: final TopModel instance after clustering and trimming, dataframe containing which run goes to which topic
    :rtype: TopModel, pandas dataframe
    """
    all_batches = None
    all_components = None
    all_exp_dirichlet_component = None
    all_others = None

    if min_cell_participation is None:
        min_cell_participation = data.shape[0] / 100

    for train in trains:
        components, exp_dirichlet_component, others = train.make_LDA_models_attributes()
        components.index = f"{train.name}_" + components.index
        exp_dirichlet_component.index = f"{train.name}_" + exp_dirichlet_component.index
        others.index = f"{train.name}_" + others.index
        batch = [train.name] * components.shape[0]
        if all_components is None:
            all_batches = batch
            all_components = components
            all_exp_dirichlet_component = exp_dirichlet_component
            all_others = others
        else:
            all_batches = batch + all_batches
            all_components = pd.concat([components, all_components], axis=0)
            all_exp_dirichlet_component = pd.concat([exp_dirichlet_component, all_exp_dirichlet_component], axis=0)
            all_others = pd.concat([others, all_others], axis=0)

    if len(trains) == 1:
        adata = anndata.AnnData(all_components)
        sc.pp.log1p(adata)
        sc.pp.highly_variable_genes(adata, n_top_genes=n_top_genes)
        sc.pp.neighbors(adata)
        sc.tl.umap(adata)
        sc.tl.leiden(adata, resolution=resolution)
        adata.obs["topics"] = adata.obs["leiden"].astype(int) + 1
        adata.obs["topics"] = "Topic_" + adata.obs["topics"].astype(str)
        sc.pl.umap(adata, color=["topics"],
                   title=[f"Topic space UMAP leiden clusters (k={trains[0].k})"],
                   save=f"_leiden_clustering.{file_format}")
        sc.pl.umap(adata, color=['topics'],
                   legend_loc='on data',
                   title=[f'Topic space UMAP leiden clusters'],
                   save=f'_leiden_clustering_v2.{file_format}')

    else:
        adata = anndata.AnnData(all_components)
        adata.obs['assays'] = all_batches
        sc.pp.log1p(adata)
        sc.pp.highly_variable_genes(adata, n_top_genes=n_top_genes)
        sc.pp.neighbors(adata)
        sce.pp.harmony_integrate(adata, 'assays', max_iter_harmony=max_iter_harmony)
        adata.obsm['X_pca'] = adata.obsm['X_pca_harmony']
        sc.pp.neighbors(adata)
        sc.tl.umap(adata)
        sc.tl.leiden(adata, resolution=resolution)
        adata.obs["topics"] = adata.obs["leiden"].astype(int) + 1
        adata.obs["topics"] = "Topic_" + adata.obs["topics"].astype(str)
        sc.pl.umap(adata, color=['topics'],
                   title=[f'Topic space UMAP leiden clusters'],
                   save=f'_leiden_clustering_harmony.{file_format}')
        sc.pl.umap(adata, color=['topics'],
                   legend_loc='on data',
                   title=[f'Topic space UMAP leiden clusters'],
                   save=f'_leiden_clustering_harmony_v2.{file_format}')
        sc.pl.umap(adata, color=['assays'],
                   title=['Topic space UMAP leiden clusters'],
                   save=f'_technology_harmony.{file_format}')

    clustering = adata.obs.copy(deep=True)
    clustering["leiden"] = clustering["leiden"].astype(int)
    n_rtopics = len(np.unique(clustering["leiden"]))

    rlda = initialize_rLDA_model(all_components,
                                 all_exp_dirichlet_component,
                                 all_others,
                                 clusters=clustering)
    lda_output = rlda.transform(data.X)
    cell_participation = pd.DataFrame(np.round(lda_output, 2),
                                      columns=[f"Topic{i + 1}" for i in range(n_rtopics)],
                                      index=data.obs.index)

    keep = cell_participation.sum() > min_cell_participation
    print(
        f"{keep.sum()} topics out of {keep.shape[0]} topics have participation more than {min_cell_participation}")
    tmp = pd.DataFrame(keep)
    tmp["cluster"] = range(tmp.shape[0])
    tmp = tmp.to_dict()
    tmp = tmp[0]
    clustering["keep"] = clustering["leiden"].astype(int) + 1
    clustering["keep"] = "Topic" + clustering["keep"].astype(str)
    clustering["keep"] = clustering["keep"].replace(tmp)
    n_rtopics = keep.sum()
    rlda, gene_weights_T = filter_LDA_model(rlda, keep)

    gene_weights = gene_weights_T.T
    gene_weights.index = data.var.index

    if len(trains) == 1:
        top_model = TopModel(name=trains[0].name,
                             N=n_rtopics,
                             gene_weights=gene_weights,
                             model=rlda)

    else:
        name = np.unique(all_batches).tolist()
        name = '_'.join(name)

        top_model = TopModel(name=name,
                             N=n_rtopics,
                             gene_weights=gene_weights,
                             model=rlda)

    return top_model, clustering, adata


def plot_cluster_contribution(clustering,
                              feature,
                              show_all=False,
                              portion=True,
                              save=True,
                              show=True,
                              file_format="pdf",
                              file_name='cluster_contribution'):
    """
    barplot shows number of topics contribute to each cluster

    :param clustering: dataframe that map each single LDA run to each cluster
    :type clustering: pandas dataframe
    :param feature: name of the feature you want to see the cluster contribution (should be one of the columns name of clustering df)
    :type feature: str
    :param show_all: Indicate if you want to show all clusters or only the ones that pass threshold (default: False)
    :type show_all: bool
    :param portion: Indicate if you want to normalized the bar to show percentage instead of actual value (default: True)
    :type portion: bool
    :param save: indicate if you want to save the plot or not (default: True)
    :type save: bool
    :param show: indicate if you want to show the plot or not (default: True)
    :type show: bool
    :param file_format: indicate the format of plot (default: pdf)
    :type file_format: str
    :param file_name: name and path of the plot use for save (default: cluster_contribution)
    :type file_name: str
    """
    if feature not in clustering.columns:
        sys.exit(f"{feature} is not valid! should be a columns names of clustering")

    options = np.unique(clustering['assays']).tolist()
    res = pd.DataFrame(columns=options + ['Topic', 'keep'],
                       index=range(len(clustering['leiden'].unique())))

    for i in clustering['leiden'].unique():
        for opt in options:
            tmp = clustering[np.logical_and(clustering['leiden'] == i,
                                            clustering['assays'] == opt)]
            res.loc[i, opt] = tmp.shape[0]

        tmp = clustering[clustering['leiden'] == i]
        res.loc[i, 'Topic'] = f"Topic_{i + 1}"
        res.loc[i, 'keep'] = tmp.keep.unique()[0]

    if not show_all:
        res = res[res["keep"]]
    res.index = res.Topic.values

    if not portion:
        plot = res.plot.bar(stacked=True,
                            xlabel='leiden',
                            ylabel='number of topics',
                            figsize=(max(res.shape[0] / 2, 5), 7))
    else:
        res['sum'] = res[options].sum(axis=1)
        for i in res.index.values:
            for opt in options:
                res.loc[i, opt] = float(res.loc[i, opt]) / res.loc[i, 'sum'] * 100

        res.drop(['Topic', 'keep', 'sum'], axis=1, inplace=True)

        plot = res.plot.bar(stacked=True,
                            xlabel='leiden',
                            ylabel='percentage(%) of topics',
                            figsize=(max(res.shape[0] / 2, 5), 7))
    fig = plot.get_figure()
    if save:
        fig.savefig(f"{file_name}.{file_format}", bbox_inches='tight')
    if show:
        plt.show()
    else:
        plt.close()


def initialize_rLDA_model(all_components, all_exp_dirichlet_component, all_others, clusters):
    """
    Initialize reproducible LDA model by calculating all necessary attributes using clustering.

    :param all_components: Variational parameters for topic gene distribution from all single LDA models
    :type all_components: pandas dataframe
    :param all_exp_dirichlet_component: Exponential value of expectation of log topic gene distribution from all single LDA models
    :type all_exp_dirichlet_component: pandas dataframe
    :param all_others: dataframe contains remaining necessary attributes including: n_batch_iter: Number of iterations of the EM step. n_features_in: Number of features seen during fit. n_iter: Number of passes over the dataset. bound: Final perplexity score on training set. doc_topic_prior: Prior of document topic distribution theta. If the value is None, it is 1 / n_components. topic_word_prior: Prior of topic word distribution beta. If the value is None, it is 1 / n_components.
    :type all_others: pandas dataframe
    :param clusters: dataframe that mapped each LDA run to each clusters
    :type clusters: pandas dataframe

    :return: Latent Dirichlet Allocation with online variational Bayes algorithm.
    :rtype: sklearn.decomposition.LatentDirichletAllocation
    """
    n_rtopics = len(np.unique(clusters[f"leiden"]))

    components = np.zeros((n_rtopics, all_components.shape[1]), dtype=float)
    exp_dirichlet_component = np.zeros((n_rtopics, all_exp_dirichlet_component.shape[1]), dtype=float)

    topic = 0
    for cluster in range(n_rtopics):
        tmp = clusters[clusters["leiden"] == cluster]

        tmp_components = all_components.loc[tmp.index, :]
        tmp_components = tmp_components.mean(axis=0)
        components[topic, :] = tmp_components.values

        tmp_exp_dirichlet_component = all_exp_dirichlet_component.loc[tmp.index, :]
        tmp_exp_dirichlet_component = tmp_exp_dirichlet_component.mean(axis=0)
        exp_dirichlet_component[topic, :] = tmp_exp_dirichlet_component.values

        topic = topic + 1

    df = pd.DataFrame(np.transpose(components),
                      columns=[f"Topic{i + 1}" for i in range(n_rtopics)],
                      index=all_components.columns)
    for topic in range(n_rtopics):
        df_sorted = df.sort_values([f"Topic{topic + 1}"], axis=0, ascending=False)[f"Topic{topic + 1}"]
        tmp = df_sorted.cumsum()
        tmp = tmp[tmp > 0.9 * df_sorted.sum()]
        df.loc[tmp.index, f"Topic{topic + 1}"] = 1 / n_rtopics

    components = df.T

    exp_dirichlet_component = pd.DataFrame(exp_dirichlet_component,
                                           index=[f"Topic{i + 1}" for i in range(n_rtopics)],
                                           columns=all_exp_dirichlet_component.columns)
    others = all_others.mean(axis=0)

    LDA = initialize_lda_model(components, exp_dirichlet_component, others)

    return LDA


def initialize_lda_model(components, exp_dirichlet_component, others):
    """
    Initialize LDA model by passing all necessary attributes

    :param components: Variational parameters for topic gene distribution
    :type components: pandas dataframe
    :param exp_dirichlet_component: Exponential value of expectation of log topic gene distribution
    :type exp_dirichlet_component: pandas dataframe
    :param others: dataframe contains remaining necessary attributes including: n_batch_iter: Number of iterations of the EM step. n_features_in: Number of features seen during fit. n_iter: Number of passes over the dataset. bound: Final perplexity score on training set. doc_topic_prior: Prior of document topic distribution theta. If the value is None, it is 1 / n_components. topic_word_prior: Prior of topic word distribution beta. If the value is None, it is 1 / n_components.
    :type others: pandas dataframe

    :return: Latent Dirichlet Allocation with online variational Bayes algorithm.
    :rtype: sklearn.decomposition.LatentDirichletAllocation
    """
    n_topics = components.shape[0]

    LDA = LatentDirichletAllocation(n_components=n_topics)

    LDA.components_ = components.values
    LDA.exp_dirichlet_component_ = exp_dirichlet_component.values
    LDA.n_batch_iter_ = int(others['n_batch_iter'])
    LDA.n_features_in_ = int(others['n_features_in'])
    LDA.n_iter_ = int(others['n_iter'])
    LDA.bound_ = others['bound']
    LDA.doc_topic_prior_ = others['doc_topic_prior']
    LDA.topic_word_prior_ = others['topic_word_prior']

    return LDA


def filter_LDA_model(main_lda, keep):
    """
    filter LDA based on the topics we want to keep

    :param main_lda: Latent Dirichlet Allocation with online variational Bayes algorithm.
    :type main_lda: sklearn.decomposition.LatentDirichletAllocation
    :param keep: dataframe that define which topics we want to keep
    :type keep: pandas dataframe

    :return: Latent Dirichlet Allocation with online variational Bayes algorithm, weights of genes in each topics (indexes are topics and columns are genes)
    :rtype: sklearn.decomposition.LatentDirichletAllocation, pandas dataframe
    """
    n_topics = keep.sum()
    lda = LatentDirichletAllocation(n_components=n_topics)

    components = main_lda.components_[[i for i in range(keep.shape[0]) if keep[i]], :]

    df = pd.DataFrame(np.transpose(components),
                      columns=[f'Topic{i + 1}' for i in range(n_topics)])
    for topic in range(n_topics):
        df_sorted = df.sort_values([f'Topic{topic + 1}'], axis=0, ascending=False)[f'Topic{topic + 1}']
        tmp = df_sorted.cumsum()
        tmp = tmp[tmp > 0.9 * df_sorted.sum()]
        df.loc[tmp.index, f'Topic{topic + 1}'] = 1 / n_topics

    components = df.T

    lda.components_ = components.values
    lda.exp_dirichlet_component_ = main_lda.exp_dirichlet_component_[[i for i in range(keep.shape[0]) if keep[i]], :]
    lda.n_batch_iter_ = main_lda.n_batch_iter_
    lda.n_features_in_ = main_lda.n_features_in_
    lda.n_iter_ = main_lda.n_iter_
    lda.bound_ = main_lda.bound_
    lda.doc_topic_prior_ = main_lda.doc_topic_prior_
    lda.topic_word_prior_ = main_lda.topic_word_prior_

    return lda, components


def read_train(file):
    """
    reading train pickle file

    :param file: path of the pickle file
    :type file: str

    :return: train instance
    :rtype: Train class
    """
    if not os.path.isfile(file):
        raise ValueError('Train file not found at given path!')
    if not file.endswith('.p') and not file.endswith('.h5'):
        raise ValueError('Train file type is not correct!')

    if file.endswith('.p'):
        picklefile = open(file, 'rb')
        train = pickle.load(picklefile)

    if file.endswith('.h5'):
        f = h5py.File(file, 'r')

        name = np.string_(f['name']).decode('ascii')
        k = np.int_(f['k'])
        n_runs = np.int_(f['n_runs'])
        random_state_range = list(f['random_state_range'])

        # models
        top_models = []
        for random_state in random_state_range:
            components = pd.DataFrame(np.array(f[f"models/{random_state}/components_"]))
            exp_dirichlet_component = pd.DataFrame(np.array(f[f"models/{random_state}/exp_dirichlet_component_"]))

            others = pd.DataFrame()
            others.loc[0, 'n_batch_iter'] = np.int_(f[f"models/{random_state}/n_batch_iter_"])
            others.loc[0, 'n_features_in'] = np.array(f[f"models/{random_state}/n_features_in_"])
            others.loc[0, 'n_iter'] = np.int_(f[f"models/{random_state}/n_iter_"])
            others.loc[0, 'bound'] = np.float_(f[f"models/{random_state}/bound_"])
            others.loc[0, 'doc_topic_prior'] = np.array(f[f"models/{random_state}/doc_topic_prior_"])
            others.loc[0, 'topic_word_prior'] = np.array(f[f"models/{random_state}/topic_word_prior_"])

            model = initialize_lda_model(components, exp_dirichlet_component, others)

            top_model = TopModel(name=f"{name}_{random_state}",
                                 N=k,
                                 gene_weights=components,
                                 model=model)
            top_models.append(top_model)

        train = Train(name=name,
                      k=k,
                      n_runs=n_runs,
                      random_state_range=random_state_range)
        train.top_models = top_models

        f.close()

    print(f"Reading Train done!")
    return train


def read_topModel(file):
    """
    reading topModel pickle/HDF5 file

    :param file: path of the pickle/HDF5 file
    :type file: str

    :return: topModel instance
    :rtype: TopModel class
    """
    if not os.path.isfile(file):
        raise ValueError('TopModel file not found at given path!')
    if not file.endswith('.p') and not file.endswith('.h5'):
        raise ValueError('TopModel file type is not correct!')

    if file.endswith('.p'):
        picklefile = open(file, 'rb')
        top_model = pickle.load(picklefile)

    if file.endswith('.h5'):
        f = h5py.File(file, 'r')

        name = np.string_(f['name']).decode('ascii')
        N = np.int_(f['N'])

        # topics
        topics = dict()
        topic_ids = [f'Topic_{i + 1}' for i in range(N)]
        for topic in topic_ids:
            topic_id = np.string_(f['topics'][topic]['id']).decode('ascii')
            topic_name = np.string_(f['topics'][topic]['name']).decode('ascii')
            gene_weights = pd.DataFrame(np.array(f['topics'][topic]['gene_weights']))
            gene_information = pd.DataFrame(np.array(f['topics'][topic]['gene_information']), dtype=str)
            gene_information.columns = gene_information.iloc[0, :]
            gene_information.drop(index=0, inplace=True)
            gene_information.index = gene_information['index']
            gene_information.drop(columns='index', inplace=True)
            gene_information.index.name = None

            topic_information = pd.DataFrame(np.array(f['topics'][topic]['topic_information']), dtype=str)
            topic_information.columns = topic_information.iloc[0, :]
            topic_information.drop(index=0, inplace=True)
            topic_information.index = topic_information['index']
            topic_information.drop(columns='index', inplace=True)
            topic_information.index.name = None

            gene_weights.index = gene_information.index.tolist()
            gene_weights.columns = topic_information.index.tolist()

            topic = Topic(topic_id=topic_id,
                          topic_name=topic_name,
                          topic_gene_weights=gene_weights,
                          gene_information=gene_information,
                          topic_information=topic_information)
            topics[topic_id] = topic

        # model
        components = pd.DataFrame(np.array(f['model']['components_']))
        exp_dirichlet_component = pd.DataFrame(np.array(f['model']['exp_dirichlet_component_']))

        others = pd.DataFrame()
        others.loc[0, 'n_batch_iter'] = np.int_(f['model']['n_batch_iter_'])
        others.loc[0, 'n_features_in'] = np.array(f['model']['n_features_in_'])
        others.loc[0, 'n_iter'] = np.int_(f['model']['n_iter_'])
        others.loc[0, 'bound'] = np.float_(f['model']['bound_'])
        others.loc[0, 'doc_topic_prior'] = np.array(f['model']['doc_topic_prior_'])
        others.loc[0, 'topic_word_prior'] = np.array(f['model']['topic_word_prior_'])

        model = initialize_lda_model(components, exp_dirichlet_component, others)

        top_model = TopModel(name=name,
                             N=N,
                             topics=topics,
                             model=model)

        f.close()

    print(f"Reading TopModel done!")
    return top_model


def read_analysis(file):
    """
    reading analysis pickle file

    :param file: path of the pickle file
    :type file: str

    :return: analysis instance
    :rtype: Analysis class
    """
    if not os.path.isfile(file):
        raise ValueError('Analysis object not found at given path!')

    picklefile = open(file, 'rb')
    analysis = pickle.load(picklefile)

    print(f"Reading Analysis done!")
    return analysis


def read_model_yaml(model_yaml_path="model.yaml",
                    topic_yaml_path=None,
                    cell_topic_participation_path=None,
                    save=True):
    """
    read YMAL files and make topmodel object
    write topic in YAML format

    :param model_yaml_path: model yaml path
    :type model_yaml_path: str
    :param topic_yaml_path: path that you use to save all topics information
    :type topic_yaml_path: str
    :param cell_topic_participation_path: path of cell-topic participation
    :type cell_topic_participation_path: str
    :param save: indicate if you want to save objects (topmodel and analysis) as a pickle file (default: True)
    :type save: bool

    :return: Topmodel and analysis objects
    :rtype: TopModel, Analysis
    """

    with open(model_yaml_path, 'r') as file:
        model_yaml = yaml.safe_load(file)

    if not all(value in list(model_yaml.keys()) for value in
               ['Topic IDs', 'Cell-Topic participation ID', 'Experiment ID', 'Name of method', 'Number of topics']):
        sys.exit("Model YMAL file is not correct!")

    topics = dict()
    topics_gene_weights = None
    for i in range(len(model_yaml['Topic IDs'])):
        topic_path = f"{topic_yaml_path}{model_yaml['Topic IDs'][i]}.yaml"
        with open(topic_path, 'r') as file:
            topic_yaml = yaml.safe_load(file)

        if not all(value in list(topic_yaml.keys()) for value in
                   ['Topic ID', 'Gene weights', 'Gene information']):
            sys.exit(f"Topic {model_yaml['Topic IDs'][i]} YMAL file is not correct!")

        topic_id = topic_yaml['Topic ID']
        topic_gene_weights = pd.DataFrame(list(topic_yaml['Gene weights'].values()),
                                          index=topic_yaml['Gene weights'].keys(),
                                          columns=[topic_yaml['Topic ID']])
        mask = topic_gene_weights.applymap(lambda x: isinstance(x, (int, float)))

        topic_gene_weights = topic_gene_weights.where(mask)
        gene_information = pd.DataFrame(topic_yaml['Gene information'])
        topic_information = None
        if 'Topic information' in topic_yaml.keys():
            topic_information = pd.DataFrame(topic_yaml['Topic information'], index=topic_gene_weights.columns)
        topics[f"Topic_{i + 1}"] = Topic(topic_id=topic_id,
                                         topic_gene_weights=topic_gene_weights,
                                         gene_information=gene_information,
                                         topic_information=topic_information)
        if topics_gene_weights is None:
            topics_gene_weights = topic_gene_weights.copy(deep=True)
        else:
            topics_gene_weights = pd.concat([topics_gene_weights, topic_gene_weights], axis=1)

    if "NMF" in model_yaml['Name of method']:
        model = NMF(n_components=int(model_yaml['Number of topics']))
    else:
        model = LatentDirichletAllocation(n_components=int(model_yaml['Number of topics']))

    model.components_ = topics_gene_weights.T.values
    topModel = TopModel(name=model_yaml['Experiment ID'],
                        N=int(model_yaml['Number of topics']),
                        topics=topics,
                        model=model)

    cell_participation = sc.read_h5ad(cell_topic_participation_path)
    analysis = Analysis(Top_model=topModel,
                        cell_participation=cell_participation)

    if save:
        topModel.save_topModel()
        analysis.save_analysis()

    return topModel, analysis


def combine_topModels(topModels,
                      name="Combined_TopModel",
                      data=None,
                      min_cell_participation=None):
    """
    Combine two topmodels. It will not apply any method when we want to combine them, so basically just combine all models without performing any method

    :param topModels: list of topmodels you want to combine
    :type topModels: list of TopModel
    :param name: name of the combined topmodels
    :type name: str
    :param data: if you want to remove topics with low cell participation, you can pass the data you used to train models
    :type data: anndata
    :param min_cell_participation: minimum cell participation across for each topics to keep them, when is None, it will keep topics with cell participation more than 1% of #cells (#cells / 100)
    :type min_cell_participation: float

    :return: return the combined TopModel, number of topics, gene weights
    :rtype: TopModel, int, pandas DataFrame
    """
    n_topics = 0
    components = None
    exp_dirichlet_component = None
    n_batch_iter = 0
    n_features_in = 0
    n_iter = 0
    bound = 0.0
    doc_topic_prior = 0.0
    topic_word_prior = 0.0
    for topmodel in topModels:
        n_topics += topmodel.N

        tmp = pd.DataFrame(topmodel.model.components_,
                           index=[f'{topmodel.name}_Topic_{i + 1}' for i in
                                  range(topmodel.model.components_.shape[0])],
                           columns=topmodel.get_feature_name())
        if components is None:
            components = tmp
        else:
            components = pd.concat([components, tmp], axis=0)

        tmp = pd.DataFrame(topmodel.model.exp_dirichlet_component_,
                           index=[f'{topmodel.name}_Topic_{i + 1}' for i in
                                  range(topmodel.model.components_.shape[0])],
                           columns=topmodel.get_feature_name())
        if exp_dirichlet_component is None:
            exp_dirichlet_component = tmp
        else:
            exp_dirichlet_component = pd.concat([exp_dirichlet_component, tmp], axis=0)

        n_batch_iter += topmodel.N * topmodel.model.n_batch_iter_
        n_features_in += topmodel.N * topmodel.model.n_features_in_
        n_iter += topmodel.N * topmodel.model.n_iter_
        bound += topmodel.N * topmodel.model.bound_
        doc_topic_prior += topmodel.N * topmodel.model.doc_topic_prior_
        topic_word_prior += topmodel.N * topmodel.model.topic_word_prior_

    n_batch_iter /= n_topics
    n_features_in /= n_topics
    n_iter /= n_topics
    bound /= n_topics
    doc_topic_prior /= n_topics
    topic_word_prior /= n_topics

    model = LatentDirichletAllocation(n_components=n_topics)

    model.components_ = components.values
    model.exp_dirichlet_component_ = exp_dirichlet_component.values
    model.n_batch_iter_ = int(n_batch_iter)
    model.n_features_in_ = int(n_features_in)
    model.n_iter_ = int(n_iter)
    model.bound_ = bound
    model.doc_topic_prior_ = doc_topic_prior
    model.topic_word_prior_ = topic_word_prior

    if data is not None:
        if min_cell_participation is None:
            min_cell_participation = data.shape[0] / 100

        lda_output = model.transform(data.X)
        cell_participation = pd.DataFrame(np.round(lda_output, 2),
                                          columns=[f"Topic{i + 1}" for i in range(n_topics)],
                                          index=data.obs.index)

        keep = cell_participation.sum() > min_cell_participation
        print(
            f"{keep.sum()} topics out of {keep.shape[0]} topics have participation more than {min_cell_participation}")
        n_topics = keep.sum()
        model, tmp = filter_LDA_model(model, keep)
        components = pd.DataFrame(model.components_,
                                  index=[f'{name}_Topic_{i + 1}' for i in
                                         range(model.components_.shape[0])],
                                  columns=components.columns.tolist())

    top_model = TopModel(name=name,
                         N=n_topics,
                         gene_weights=components.T,
                         model=model)

    return top_model, n_topics, components.T



import sys
import warnings
import joblib
import pickle
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import h5py

sns.set_context('paper')
warnings.filterwarnings('ignore')

from Topyfic.topic import Topic
from Topyfic.utilsAnalyseModel import MA_plot


class TopModel:
    """
    A class that saved a model

    :param name: name of class
    :type name: str
    :param N: number of topics
    :type N: int
    :param gene_weights: dataframe that has weights of genes for each topics; genes are indexes and topics are columns
    :type gene_weights: pandas dataframe
    :param topics: dictionary contains all topics for the topmodel
    :type topics: Dictionary of Topics
    :param model: store reproducible LDA model
    :type model: sklearn.decomposition.LatentDirichletAllocation

    """

    def __init__(self,
                 name,
                 N,
                 topics=None,
                 gene_weights=None,
                 gene_information=None,
                 model=None):
        if gene_weights is None and topics is None:
            sys.exit("Both gene weights and topics can not be empty at the same time!")

        if topics is not None:
            self.topics = topics
        else:
            self.topics = dict()
            for i in range(N):
                topic_gene_weights = pd.DataFrame(gene_weights.iloc[:, i].values,
                                                  index=gene_weights.index,
                                                  columns=[f"Topic_{i + 1}"])
                self.topics[f"Topic_{i + 1}"] = Topic(topic_id=f"Topic_{i + 1}",
                                                      topic_gene_weights=topic_gene_weights,
                                                      gene_information=gene_information)
        self.name = name
        self.N = N
        self.model = model

    def get_feature_name(self):
        """
        get feature(gene) name

        :return: list of feature(gene) name
        :rtype: list
        """
        topic1 = next(iter(self.topics.items()))[1]
        return topic1.gene_information.index.tolist()

    def save_rLDA_model(self, name='rLDA', save_path="", file_format='joblib'):
        """
            save rLDA model (instance of LDA model in sklearn) as a joblib/HDF5 file.

            :param name: name of the joblib file (default: rLDA)
            :type name: str
            :param save_path: directory you want to use to save pickle file (default is saving near script)
            :type save_path: str
            :param file_format: format of the file you want to save (option: joblib (default), HDF5)
            :type file_format: str
        """
        if file_format not in ['joblib', 'HDF5']:
            sys.exit(f"{file_format} is not correct! It should be 'joblib' or 'HDF5'.")

        if file_format == "joblib":
            print(f"Saving rLDA model as {name}_{self.N}topics.joblib")

            joblib.dump(self.model, f"{save_path}{name}_{self.N}topics.joblib", compress=3)

        if file_format == "HDF5":
            print(f"Saving rLDA model as {name}_{self.N}topics.h5")

            f = h5py.File(f"{name}_{self.N}topics.h5", "a")

            f['components_'] = self.model.components_
            f['exp_dirichlet_component_'] = self.model.exp_dirichlet_component_
            f['n_batch_iter_'] = np.int_(self.model.n_batch_iter_)
            f['n_features_in_'] = self.model.n_features_in_
            f['n_iter_'] = np.int_(self.model.n_iter_)
            f['bound_'] = np.float_(self.model.bound_)
            f['doc_topic_prior_'] = np.float_(self.model.doc_topic_prior_)
            f['topic_word_prior_'] = np.float_(self.model.topic_word_prior_)

            f.close()

    def get_gene_weights(self):
        """
        get feature(gene) weights

        :return: dataframe contains feature(gene) weights; genes are indexes and topics are columns
        :rtype: pandas dataframe
        """
        gene_weights = pd.DataFrame(np.transpose(self.model.components_),
                                    columns=[f'{self.name}_Topic_{i + 1}' for i in
                                             range(self.model.components_.shape[0])],
                                    index=self.get_feature_name())

        return gene_weights

    def get_ranked_gene_weight(self):
        """
        get sorted feature(gene) weights. each value is gene and weights on each topics

        :return: dataframe contains feature(gene) and their weights; ranks are indexes and topics are columns
        :rtype: pandas dataframe
        """
        gene_weights = pd.DataFrame(np.transpose(self.model.components_),
                                    columns=[f'{self.name}_Topic_{i + 1}' for i in
                                             range(self.model.components_.shape[0])],
                                    index=self.get_feature_name())
        ranked_gene_weights = pd.DataFrame(columns=[f'{self.name}_Topic_{i + 1}' for i in
                                                    range(self.model.components_.shape[0])],
                                           index=range(len(self.get_feature_name())))
        for col in gene_weights.columns:
            tmp = gene_weights.sort_values(by=col, ascending=False)
            tmp[col] = ";" + tmp[col].astype(str)
            tmp[col] = tmp.index.tolist() + tmp[col].astype(str)
            ranked_gene_weights[col] = tmp[col].values
        return ranked_gene_weights

    def get_top_model_attributes(self):
        """
        get top model attributes to be able to make sklearn.decomposition.LatentDirichletAllocation


        :return: three data frame which the first one is components, the second one is exp_dirichlet_component and
        the last one is combining the rest of LDA attributes which put them to gather as a dataframe
        :rtype: pandas dataframe, pandas dataframe, pandas dataframe
        """
        feature = self.get_feature_name()

        components = pd.DataFrame(self.model.components_,
                                  index=[f"Topic_{i + 1}" for i in range(self.N)],
                                  columns=feature)

        exp_dirichlet_component = pd.DataFrame(self.model.exp_dirichlet_component_,
                                               index=[f"Topic_{i + 1}" for i in range(self.N)],
                                               columns=feature)

        others = pd.DataFrame(
            index=[f"Topic_{i + 1}" for i in range(self.N)],
            columns=["n_batch_iter",
                     "n_features_in",
                     "n_iter",
                     "bound",
                     "doc_topic_prior",
                     "topic_word_prior"])

        others.loc[[f"Topic_{i + 1}" for i in range(self.N)], "n_batch_iter"] = self.model.n_batch_iter_
        others.loc[[f"Topic_{i + 1}" for i in range(self.N)], "n_features_in"] = self.model.n_features_in_
        others.loc[[f"Topic_{i + 1}" for i in range(self.N)], "n_iter"] = self.model.n_iter_
        others.loc[[f"Topic_{i + 1}" for i in range(self.N)], "bound"] = self.model.bound_
        others.loc[[f"Topic_{i + 1}" for i in range(self.N)], "doc_topic_prior"] = self.model.doc_topic_prior_
        others.loc[[f"Topic_{i + 1}" for i in range(self.N)], "topic_word_prior"] = self.model.topic_word_prior_

        return components, exp_dirichlet_component, others

    def gene_weight_rank_heatmap(self,
                                 genes=None,
                                 topics=None,
                                 show_rank=True,
                                 scale=None,
                                 save=True,
                                 show=True,
                                 figsize=None,
                                 file_format="pdf",
                                 file_name="gene_weight_rank_heatmap"):
        """
        plot selected genes weights and their ranks in selected topics

        :param genes: list of genes you want to see their weights (default: all genes)
        :type genes: list
        :param topics: list of topics
        :type topics: list
        :param show_rank: indicate if you want to show the rank of significant genes or not (default: True)
        :type show_rank: bool
        :param scale: indicate if you want to plot as log2, log10 or not (default: None which show actual value) other options is log2 and log10
        :scale scale: str
        :param save: indicate if you want to save the plot or not (default: True)
        :type save: bool
        :param show: indicate if you want to show the plot or not (default: True)
        :type show: bool
        :param figsize: indicate the size of plot (default: (10 * (len(category) + 1), 10))
        :type figsize: tuple of int
        :param file_format: indicate the format of plot (default: pdf)
        :type file_format: str
        :param file_name: name and path of the plot use for save (default: gene_weight_rank_heatmap)
        :type file_name: str
        """
        if genes is None:
            genes = self.get_feature_name()
        elif not (set(genes) & set(self.get_feature_name())) == set(genes):
            sys.exit("some/all of genes are not part of topModel!")

        if topics is None:
            topics = [f'{self.name}_Topic_{i + 1}' for i in range(self.model.components_.shape[0])]
        elif not (set(topics) & set(
                [f'{self.name}_Topic_{i + 1}' for i in range(self.model.components_.shape[0])])) == set(topics):
            sys.exit("some/all of topics are not part of topModel!")

        if scale not in ["log2", "log10", None]:
            sys.exit("scale is not valid!")

        if figsize is None:
            figsize = (self.N * 1.5, len(genes))

        gene_weights = self.get_gene_weights()
        gene_weights = gene_weights.loc[:, topics]
        gene_rank = gene_weights.loc[genes, topics]
        for topic in topics:
            tmp = gene_weights.sort_values([topic], ascending=False)
            tmp['rank'] = range(1, tmp.shape[0] + 1)
            gene_rank.loc[genes, topic] = tmp.loc[genes, 'rank'].values

        gene_weights = self.get_gene_weights()
        gene_weights = gene_weights.loc[genes, topics]

        gene_weights = gene_weights.reindex(genes)
        gene_weights = gene_weights.reindex(topics, axis="columns")

        gene_rank = gene_rank.reindex(genes)
        gene_rank = gene_rank.reindex(topics, axis="columns")
        gene_rank = gene_rank.astype(int)
        gene_rank[gene_weights <= self.N] = 0
        gene_rank = gene_rank.astype(str)
        gene_rank[gene_rank == "0"] = ""

        fig, ax = plt.subplots(figsize=figsize,
                               facecolor='white')

        if scale == "log2":
            gene_weights = gene_weights.applymap(np.log2)
        if scale == "log10":
            gene_weights = gene_weights.applymap(np.log10)
        if scale is None:
            scale = "no"

        if show_rank:
            sns.heatmap(gene_weights,
                        cmap='viridis',
                        annot=gene_rank,
                        linewidths=.5,
                        fmt='',
                        ax=ax)
            ax.set_title(f'gene weights with rank ({scale} scale)')
        else:
            sns.heatmap(gene_weights,
                        cmap='viridis',
                        linewidths=.5,
                        ax=ax)
            ax.set_title('gene weights')

        if save:
            plt.savefig(f"{file_name}.{file_format}")
        if show:
            plt.show()
        else:
            plt.close()

    def MA_plot(self,
                topic1,
                topic2,
                size=None,
                pseudocount=1,
                threshold=1,
                cutoff=2,
                consistency_correction=1.4826,
                #topN=None,
                labels=None,
                save=True,
                show=True,
                file_format="pdf",
                file_name="MA_plot"):
        """
        plot MA based on the gene weights on given topics

        :param topic1: first topic to be compared
        :type topic1: str
        :param topic2: second topic to be compared
        :type topic2: str
        :param size: table contains size of dot for each genes (genes are index)
        :type size: pandas dataframe
        :param pseudocount: pseudocount that you want to add (default: 1)
        :type pseudocount: float
        :param threshold: threshold to filter genes based on A values (default: 1)
        :type threshold: float
        :param cutoff: cutoff for categorized genes by modified z-score (default: 2)
        :type cutoff: float
        :param consistency_correction: the factor converts the MAD to the standard deviation for a given distribution. The default value (1.4826) is the conversion factor if the underlying data is normally distributed
        :type consistency_correction: float
        :param topN: number of genes to be consider for calculating z-score based on the A value (if it's none is gonna be avarage of # genes in both topics with weights above threshold
        :type topN: int
        :param labels: list of gene names wish to show in MA-plot
        :type labels: list
        :param save: indicate if you want to save the plot or not (default: True)
        :type save: bool
        :param show: indicate if you want to show the plot or not (default: True)
        :type show: bool
        :param file_format: indicate the format of plot (default: pdf)
        :type file_format: str
        :param file_name: name and path of the plot use for save (default: MA_plot)
        :type file_name: str

        :return: return M and A values
        """
        gene_weights = self.get_gene_weights().copy(deep=True)
        topic1 = gene_weights[topic1]
        topic2 = gene_weights[topic2]

        gene_zscore = MA_plot(topic1,
                              topic2,
                              size=size,
                              pseudocount=pseudocount,
                              threshold=threshold,
                              cutoff=cutoff,
                              consistency_correction=consistency_correction,
                              #topN=topN,
                              labels=labels,
                              save=save,
                              show=show,
                              file_format=file_format,
                              file_name=file_name)

        return gene_zscore

    def save_topModel(self, name=None, save_path="", file_format='pickle'):
        """
            save TopModel class as a pickle/HDF5 file

            :param name: name of the file (default: topModel_TopModel.name)
            :type name: str
            :param save_path: directory you want to use to save pickle file (default is saving near script)
            :type save_path: str
            :param file_format: format of the file you want to save (option: pickle (default), HDF5)
            :type file_format: str
        """
        if file_format not in ['pickle', 'HDF5']:
            sys.exit(f"{file_format} is not correct! It should be 'pickle' or 'HDF5'.")
        if name is None:
            name = f"topModel_{self.name}"

        if file_format == "pickle":
            print(f"Saving topModel as {name}.p")

            picklefile = open(f"{save_path}{name}.p", "wb")
            pickle.dump(self, picklefile)
            picklefile.close()

        if file_format == "HDF5":
            print(f"Saving topModel as {name}.h5")

            f = h5py.File(f"{name}.h5", "w")
            # model
            model = f.create_group("model")
            model['components_'] = self.model.components_
            model['exp_dirichlet_component_'] = self.model.exp_dirichlet_component_
            model['n_batch_iter_'] = np.int_(self.model.n_batch_iter_)
            model['n_features_in_'] = self.model.n_features_in_
            model['n_iter_'] = np.int_(self.model.n_iter_)
            model['bound_'] = np.float_(self.model.bound_)
            model['doc_topic_prior_'] = np.float_(self.model.doc_topic_prior_)
            model['topic_word_prior_'] = np.float_(self.model.topic_word_prior_)

            # topics
            topics = f.create_group("topics")
            for topic in self.topics.keys():
                topic_gp = topics.create_group(self.topics[topic].id)
                topic_gp['id'] = np.string_(self.topics[topic].id)
                topic_gp['name'] = np.string_(self.topics[topic].name)
                topic_gp['gene_weights'] = self.topics[topic].gene_weights
                gene_information = self.topics[topic].gene_information.copy(deep=True)
                gene_information.reset_index(inplace=True)
                gene_information = gene_information.T.reset_index().T
                topic_gp['gene_information'] = np.array(gene_information)
                topic_information = self.topics[topic].topic_information.copy(deep=True)
                topic_information.reset_index(inplace=True)
                topic_information = topic_information.T.reset_index().T
                topic_gp['topic_information'] = np.array(topic_information)

            f['name'] = np.string_(self.name)
            f['N'] = np.int_(self.N)

            f.close()




from setuptools import setup

setup(
    name='Topyfic',  # the name of your package
    packages=['Topyfic'],  # same as above
    version='v0.4.18',  # version number
    license='MIT',  # license type
    description='Topyfic is a Python package designed to identify reproducible latent dirichlet allocation (LDA) '
                'using leiden clustering and harmony for single cell epigenomics data',
    # short description
    author='Narges Rezaie',  # your name
    author_email='nargesrezaie80@gmail.com',  # your email
    url='https://github.com/mortazavilab/Topyfic',  # url to your git repo
    download_url='https://github.com/mortazavilab/Topyfic/archive/refs/tags/v0.4.18.tar.gz',  # link to the tar.gz file associated with this release
    keywords=['Cellular Programs', 'Latent Dirichlet allocation', 'single-cell multiome', 'single-cell RNA-seq',
              'gene regulatory network', 'Topic Modeling', 'single-nucleus RNA-seq'],  #
    python_requires='>=3.9',
    install_requires=[  # these can also include >, <, == to enforce version compatibility
        'pandas>=1.4.4',  # make sure the packages you put here are those NOT included in the base python distribution
        'scikit-learn>=0.24.2',
        'pytest',
        'leidenalg',
        'setuptools',
        'anndata>=0.9.2',
        'click>=8.0.4',
        'gseapy>=1.0.5',
        'joblib>=1.3.2',
        'matplotlib>=3.7.2',
        'networkx>=2.8.4',
        'numpy>=1.24.4',
        'reactome2py>=3.0.0',
        'scanpy>=1.9.5',
        'scikit_learn>=1.3.0',
        'scipy>=1.11.2',
        'seaborn>=0.11.2',
        'statsmodels>=0.14.0',
        'adjustText>=0.8',
        'PyYAML>=6.0.1',
        'click>=8.1.7',
        'umap',
        'obonet',
        'plotly',
        'h5py',

    ],
    classifiers=[  # choose from here: https://pypi.org/classifiers/
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research ',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.10',
    ],
)
